unit binStreamsU;

{
  Unit to handle binary streams.
  There will be < 500 binary streams. Therefore a simple linked list will
  suffice. The added complexity of an ordered list or tree structure will
  increase the complexity of the code but will not produce a significant
  performance increase.
}

interface

  type

    TbinStreams = ^TbinStream;

    TbinStream = record
      variable: string;
      value: string;
      next: TbinStreams;
    end;


  function emptyBinStreams: TbinStreams;
  procedure fetchStreamsFromFile(inFilename: String; var binStreams: TbinStreams);
  procedure Lookup(binStreams: TbinStreams; S: string);
  function lookupBit(binStreams: TbinStreams; S: string; bit: integer): boolean;
  function longestBinStream(binStreams: TbinStreams): integer;
  procedure addBinStreams(var binStreams: TbinStreams; newBinStream: TBinStream);
  function populateTbinStream(declaration: string): TbinStream;
  procedure putResultsToFile(outFilename: string; results: TbinStreams);



implementation

const
  { unitName = 'binStreamsU'; { For debugging }
  unitName = ''; { Turn off debugging messages }


{ Function to return an empty list of TbinStreams }
function emptyBinStreams: TbinStreams;
  begin
    result := nil;
  end;


{ Procedure to add a new TbinStream to the front of a TbinStreams }
procedure addBinStreams(var binStreams: TbinStreams; newBinStream: TBinStream);
  var
    newBinStreams: TbinStreams;

  begin
    new(newBinStreams);
    newBinStreams^.variable := newBinStream.variable;
    newBinStreams^.value := newBinStream.value;
    newBinStreams^.next := binStreams;
    binStreams := newBinStreams;
  end;


{ Procedure to print all the binStreams in a TbinStreams to stdout }
procedure printBinStreams(binStreams: TbinStreams);
  begin
    while (binStreams <> nil) do begin
      write('<' + binStreams^.variable + '> = ');
      writeln('<' + binStreams^.value + '>');
      binStreams := binStreams^.next;
    end;
  end;


{ Function to return a TbinStream from a declaration of the form <var>=<value> }
function populateTbinStream(declaration: string): TbinStream;
  var
    counter: integer; { Counter }

  begin
    { Variable initialisation }
    counter := 1;
    result.variable := ''; { Variable name that is being declared }
    result.value := ''; { Value of the variable that is being declared }
    result.next := emptyBinStreams();
    { End of variable initialisation }

    while ((counter <= Length(declaration)) AND (declaration[counter] <> '=')) do begin
      if (declaration[counter] <> ' ') then { strip out space, even in variable names }
        result.variable := result.variable + declaration[counter];
        counter := counter + 1;
    end;
    counter := counter + 1; { skip the equals sign }
    while (counter <= Length(declaration)) do begin
      if (declaration[counter] <> ' ') then { strip out space, even in variable names }
        result.value := result.value + declaration[counter];
      counter := counter + 1;
    end;

  end;


{ Procedure to lookup a variable in binStreams and print its value to the screen }
procedure Lookup(binStreams: TbinStreams; S: string);
  var
    flag: boolean;

  begin
    { Variable initialisation }
    flag := TRUE;
    { End of variable initialisation }

    while ((binStreams <> nil) AND FLAG) do begin
      if binStreams^.variable = S then begin
        writeln('  <' + binStreams^.variable + '> = <' + binStreams^.value + '>');
        writeln;
        flag := FALSE;
      end
      else
        binStreams := binStreams.next;
    end;
    if FLAG then
      writeln('  ERROR: Variable not found!');
      writeln;
  end;


{ Procedure to lookup a variable in binStreams and return the specified bit }
function lookupBit(binStreams: TbinStreams; S: string; bit: integer): boolean;
  var
    flag: boolean;

  begin
    { Variable initialisation }
    flag := TRUE;
    { End of variable initialisation }

    while ((binStreams <> emptyBinStreams) AND FLAG) do begin
      if binStreams^.variable = S then begin
        if (bit <= length(binStreams^.value)) then
          if (binStreams^.value[bit] >= '1') then
            result := TRUE
          else
            result := FALSE
        else
          result := FALSE;

        { writeln('  <' + binStreams^.variable + '> = <' + binStreams^.value[bit] + '>'); }
        flag := FALSE;
      end
      else
        binStreams := binStreams.next;
      end;
    if FLAG then begin
      result := FALSE;
      writeln('  ERROR: ' + S + ' not found!');
    end;

  end;


function longestBinStream(binStreams: TbinStreams): integer;
  begin
    result := 0; { Initialise result }
    while (binStreams <> nil) do begin
      if (length(binStreams^.value) > result) then
        result := length(binStreams^.value);
      binStreams := binStreams.next;
    end;

  end;


procedure putResultsToFile(outFilename: string; results: TbinStreams);
  var
    OutFile: TextFile; { File handle }

  begin
    { Variable initialisation }
    AssignFile(OutFile, outFilename); { Set up the file handle }
    { End of variable initialisation }

    if (unitName <> '') then writeln(unitName + '.putResultsToFile: Writing to ' + outFilename + '...');
    Rewrite(OutFile);
    while (results <> emptyBinStreams) do begin
      writeln(OutFile, results^.variable + '=' + results^.value);
      results := results^.next;
    end;
    Close(OutFile);
    if (unitName <> '') then writeln(unitName + '.putResultsToFile: Succesfully wrote ' + outFilename + '...');
    if (unitName <> '') then writeln;

  end;



{ Main Program }
procedure fetchStreamsFromFile(inFilename: String; var binStreams: TbinStreams);
  var
    InFile: TextFile; { File handle }
    tmpString: string; { Temporary string }

  begin
    { Variable initialisation }
    binStreams := emptyBinStreams(); { Make an empty set of binStreams }
    AssignFile(InFile, inFilename); { Set up the file handle }
    { End of variable initialisation }

    if (unitName <> '') then writeln(unitName + '.fetchStreamsFromFile: Reading from ' + inFilename + '...');
    Reset(InFile); { Open the file }
    while NOT(EOF(InFile)) do begin { Process until EOF }
      readln(InFile, tmpString); { Read a line in }
      addBinStreams(binStreams, populateTbinStream(tmpString)); { Put the variable in the list }
    end;
    Close(InFile); { Close the file }
    if (unitName <> '') then writeln(unitName + '.fetchStreamsFromFile: Succesfully read ' + inFilename + '...');
    if (unitName <> '') then writeln;
end;



end.
 