unit parseTreeU;

{
  Unit to handle parse trees
}

interface

uses
  binStreamsU;

  type

    Toperator = (EQUALS, LOGICALOR, LOGICALAND, LOGICALNOT);

    Texpression = ^TExpNode;

    TexpNode = record
      case Leaf : boolean of
        TRUE :
          ( variable : String[5] );

        FALSE :
          ( operand1 : Texpression;
            operator : Toperator;
            operand2 : Texpression );
    end;

    TparseTrees = ^TparseTree;

    TparseTree = record
      parseTree: Texpression;
      next: TparseTrees;

  end;

  procedure fetchFunctionsFromFile(inFilename: String; var parseTrees: TparseTrees);
  function evaluateFunctions(binStreams: TbinStreams; parseTrees: TparseTrees; var results: TbinStreams): TbinStreams;
  function emptyParseTrees: TparseTrees;

{ -----------------------------------------------------------------------------}



implementation

uses
  lexemeListU;

const
  { unitName = 'parseTreeU'; { For debugging }
  unitName = ''; { Disable debugging messages }

  { Prototypes for access procedures for parseTrees }
  function buildLeaf(variable: String): Texpression;                                     forward;
  function buildTree(operand1, operand2: Texpression; operator: Toperator): Texpression; forward;
  procedure parseDefinition(var lexList: Tlexemes; var tree: Texpression);               forward;
  procedure parseExpression(var lexlist: Tlexemes; var tree: Texpression);               forward;
  procedure parseTerm(var lexList: Tlexemes; var tree: Texpression);                     forward;
  procedure parseFactor(var lexList: Tlexemes; var tree: Texpression);                   forward;
  function emptyTree: Texpression;                                                       forward;
  function evaluate(tree: Texpression; binStreams: TbinStreams; bit: integer): boolean;  forward;
  procedure addParseTree(var parseTrees: TparseTrees; newParseTree: Texpression);        forward;
  procedure reverseParseTrees(var parseTrees : TparseTrees);                             forward;

{ -----------------------------------------------------------------------------}



{ Function definitions for parseTrees }
{ Function to build the leaf of a tree }
function buildLeaf(variable: String): Texpression;
  begin
    new(result);
    result^.leaf := TRUE;
    result^.variable := variable;
  end;


{ Function to build a tree }
function buildTree(operand1, operand2: Texpression; operator: Toperator): Texpression;
  begin
    new(result);
    result^.Leaf := FALSE;
    result^.operand1 := operand1;
    result^.operand2 := operand2;
    result^.operator := operator;
  end;


{ Procedure to parse a definition }
procedure parseDefinition(var lexList: Tlexemes; var tree: Texpression);
  var
    operator: string;
    operand: Texpression;

  begin
    if (unitName <> '') then writeln(unitName + '.parseDefinition: Parsing definition...');
    parseExpression(lexList, tree);
    while ((lexList <> emptyLexemes) AND (firstLexeme(lexList) = '=')) do begin
      if (unitName <> '') then writeln('Defn: ' + lexList^.lexeme);
      operator := firstLexeme(lexList);
      lexList := restLexemes(lexList);
      parseExpression(lexList, operand);
      if operator = '=' then
        tree := buildTree(tree, operand, EQUALS)
    end;

  end;


{ Procedure to parse an expression }
procedure parseExpression(var lexList: Tlexemes; var tree: Texpression);
  var
    operator: string;
    operand: Texpression;

  begin
    if (unitName <> '') then writeln(unitName + '.parseExpression: Parsing expression...');
    parseTerm(lexList, tree);
    while ((lexList <> emptyLexemes) AND (firstLexeme(lexList) = 'OR')) do begin
      if (unitName <> '') then writeln('Expn: ' + lexList^.lexeme);
      operator := firstLexeme(lexList);
      lexList := restLexemes(lexList);
      parseTerm(lexList, operand);
      if operator = 'OR' then
        tree := buildTree(tree, operand, LOGICALOR)
    end;
  end;


{ Procedure to parse a term }
procedure parseTerm(var lexList: Tlexemes; var tree: Texpression);
  var
    operator: string;
    operand: Texpression;

  begin
    if (unitName <> '') then writeln(unitName + '.parseTerm: Parsing term...');
    parseFactor(lexList, tree);
    while ((lexList <> emptyLexemes) AND (firstLexeme(lexList) = 'AND')) do begin
      if (unitName <> '') then writeln('Term: ' + lexList^.lexeme);
      operator := firstLexeme(lexList);
      lexList := restLexemes(lexList);
      parseFactor(lexList, operand);
      if operator = 'AND' then
        tree := buildTree(tree, operand, LOGICALAND);
    end;
  end;


{ Procedure to parse a factor }
procedure parseFactor(var lexList: Tlexemes; var tree: Texpression);
  var
    lexeme : string;

  begin
    if (unitName <> '') then writeln(unitName + '.parseFactor: Parsing factor...');
    if (unitName <> '') then writeln('Fact: ' + lexList^.lexeme);
    lexeme := firstLexeme(lexList);
    lexList := restLexemes(lexList);
    { writeln('<' + lexeme + '>'); }

    if (lexeme <> '(') AND (lexeme <> 'NOT(') then
      tree := buildLeaf(lexeme)
    else begin
      parseExpression(lexList, tree);
      if lexeme = 'NOT(' then tree := buildTree(tree, nil, LOGICALNOT);
      {Read past the right bracket}
      lexList := restLexemes(lexList);
    end;

  end;


{ Function to return an empty tree }
function emptyTree: Texpression;
  begin
    result := nil;
  end;


{ Function to evaluate the result of a parse tree }
function evaluate(tree: Texpression; binStreams: TbinStreams; bit: integer): boolean;
  begin
    if (unitName <> '') then writeln(unitName + '.evaluate: evaluating...');
    if tree^.leaf then begin
      if (unitName <> '') then writeln(unitName + '.evaluate: this is a leaf...');
      evaluate := lookupBit(binStreams, tree^.variable, bit);
    end
    else if tree^.operator = LOGICALOR then
      evaluate := evaluate(tree^.operand1, binStreams, bit) OR evaluate(tree^.operand2, binStreams, bit)
    else if tree^.operator = LOGICALAND then
      evaluate := evaluate(tree^.operand1, binStreams, bit) AND evaluate(tree^.operand2, binStreams, bit)
    else if tree^.operator = EQUALS then
      evaluate := evaluate(tree^.operand2, binStreams, bit)
    else if tree^.operator = LOGICALNOT then
      evaluate := NOT(evaluate(tree^.operand1, binStreams, bit));
end;


{ Procedure to add a new TparseTree to the front of a TparseTrees }
procedure addParseTree(var parseTrees: TparseTrees; newParseTree: Texpression);
  var
    newParseTrees: TparseTrees;

  begin
    new(newParseTrees);
    newParseTrees^.parseTree := newParseTree;
    newParseTrees^.next := parseTrees;
    parseTrees := newParseTrees;

    if (unitName <> '') then writeln(unitName + '.addParseTree: Adding a parse tree to list...');

  end;


{ Function that returns an empty list of parse trees }
function emptyParseTrees: TparseTrees;
  begin
    result := nil;
  end;


{ Procedure to reverse a list of parse trees }
procedure reverseParseTrees(var parseTrees : TparseTrees);
  var
    tmpParseTrees: TparseTrees;

  begin
    tmpParseTrees := emptyParseTrees;
    while parseTrees <> emptyParseTrees do begin
      addParseTree(tmpParseTrees, parseTrees^.parseTree);
      parseTrees := parseTrees^.next;
    end;
    parseTrees := tmpParseTrees;
  end;


{ -----------------------------------------------------------------------------}



{ Main Program }
procedure fetchFunctionsFromFile(inFilename: String; var parseTrees: TparseTrees);
  var
    InFile: TextFile; { File handle }
    tmpString: string; { Temporary string }
    lexemes: Tlexemes;
    parseTree: Texpression;

  begin
    { Variable initialisation }
    lexemes := emptyLexemes(); { Make an empty set of lexemes }
    AssignFile(InFile, inFilename); { Set up the file handle }
    { End of variable initialisation }

    if (unitName <> '') then writeln(unitName + '.fetchFunctionsFromFile: Reading from ' + inFilename + '...');

    Reset(InFile); { Open the file }
    while NOT(EOF(InFile)) do begin { Process until EOF }
      new(parseTree);  { Allocate some memory for the parse tree about to be read }
      parseTree := emptyTree; { Make the new parse tree empty }
      readln(InFile, tmpString); { Read a line in to tmpString }
      deduceLexemes(lexemes, tmpString); { Put the lexemes in tmpString in the list lexemes }
      reverseLexemes(lexemes); { Reverse the list of lexemes because it turns out backwards }
      parseDefinition(lexemes, parseTree); { Turn the list of lexemes into a parse tree }
      addParseTree(parseTrees, parseTree); { Add this parse tree to the list of parse trees }
      destroyLexemes(lexemes); { Destroy the list of lexemes }
    end;
    Close(InFile); { Close the file }
    reverseParseTrees(parseTrees);
    if (unitName <> '') then writeln(unitName + '.fetchFunctionsFromFile: Succesfully read ' + inFilename + '...');
    if (unitName <> '') then writeln;
end;


function evaluateFunctions(binStreams: TbinStreams; parseTrees: TparseTrees; var results: TbinStreams): TbinStreams;
  var
    counter: integer;
    tmpString: String;

  begin
    while (parseTrees <> emptyParseTrees) do begin { Process each parse tree in the list }
      counter := 1;
      tmpString := parseTrees^.parseTree^.operand1^.variable + '='; { Initialise our answer string to 'Var=' }
      while (counter <= longestBinStream(binStreams)) do begin { Evaluate the parse tree for each bit in the bit stream }
        if (evaluate(parseTrees^.parseTree, binStreams, counter)) then
          tmpString := tmpString + '1' { Add result to the answer string }
        else
          tmpString := tmpString + '0'; { Add result to the answer string }

        counter := counter + 1;
      end;
      addBinStreams(result, populateTbinStream(tmpString)); { Put the variable in the list }
      addBinStreams(binStreams, populateTbinStream(tmpString)); { Make the values already calculated available for future use }
      parseTrees := parseTrees.next;
    end;

  end;

{ -----------------------------------------------------------------------------}



end.
